%{
/**************************************
*  The bison file for IA32 instruction Length
*  decoder
**************************************
**************************************
*  Written by
*  Shaun Xiushan Feng
*  April, 2005
********* *****************************/

/* Reference: The IA-32 Intel Architecture Software Developer’s Manual,
   2004. Four volumes. Intel Order Numbers 253665–253668.
*/
                                                                                
	#include<iostream>
    	#include <vector>
	#include "vector_perm.h"
    	#include <string>
    	#include <stdio.h>
    	#include <stdlib.h>
   	#include <string.h> 	
	#define YYERROR_VERBOSE

	#define false           0
	#define true            !(false)
	#define INIT_ADDRESS_MODE true
	#define INIT_OPRAND_MODE true
                                                                                         
	#define PreF0h                  "11110000"
	#define PreF2h                  "11110010"
	#define PreF3h                  "11110011"
	#define Pre66h                  "01100110"
	#define Pre67h                  "01100111"
    	using namespace std;

    	extern FILE *yyin;
    	extern char *yytext;
    	extern void lex_init();
    	extern int yylex();

    	struct mod_ {
		mod_(){
			pattern = "";
			size = 0;
		}
		string pattern;
		int size;
    	};

        void print_all(string prefix_string);
	void init_prefix();
        void init_instruction();
	void mod_16_bit_address(int mod_11, string ext);
        void mod_32_bit_address(int mod_11, string ext);
	string removeBlanks(string bits);

        /* NOTE I only handle Prefix 0x66h and 0x67 which can change the length
           of instructions here, but you can easily add other prefixes into 
           their groups
        */
	string Pre_gr1[3] = {"", "", ""};
	string Pre_gr2[8] = {"", "", "", "", "", "", "", ""};
	string Pre_gr3[1] = {Pre66h};
	string Pre_gr4[1] = {Pre67h}; 	
	int gr1 = false;
        int gr2 = false;
        int gr3 = false;
        int gr4 = false;

    	int yyerror(char *s)
     	{
      		fprintf(stderr, "%s\n", s);
      		return 0;
     	}
	/* default 32_bit addressing mode */
     	int current_address_32 = INIT_ADDRESS_MODE; 
	/* default 32_bit oprand mode */
     	int current_oprand_32  = INIT_OPRAND_MODE; 

	int current_length = 0;
	int immediate_flag = false;
	string current_ID;
     	string current_string;

     	vector<string> prefix_array;
	vector< vector<string> > prefix_array_combination;
	vector<struct mod_> mod_16_array, mod_32_array;
	
%}
%token IMMEDIATE R_M REG MOD MODA BIT_8 BIT_3 BIT_2
%token MANDATORY_PREFIX ESCAPE IMM8 IMM16 ID
%start instruction_list
%union  yystacktype     {
        char *string;
	int integer;
}
%type <string> BIT_8 BIT_3 BIT_2 MANDATORY_PREFIX ESCAPE ID
%type <string> mandatory escape opcode 

%%

instruction : ID {
	      current_ID = (string)$1;
	      free($1);
              /* List instructions using Instruction Format II Table B-11, Intel */
            }
              mandatory {
              string man;
              if ($3!= NULL) {
                  man = removeBlanks((string)$3);
		  if (man == PreF2h) gr1 = true;
		  if (man == PreF3h) gr1 = true;
		  if (man == Pre66h) gr3 = true;
                  current_string
                   = current_string + man +":";
                  free($3);
              }
            } escape {
              if ($5!= NULL) {
                  current_string
                   = current_string +  removeBlanks((string)$5) + ":";
                  free($5);
              } 
            } opcode {
                /* need to replace s here*/
                char *rep;
                if (rep = strchr($7, 's')) *rep = 'x';
		current_string
                   = current_string + removeBlanks((string)$7);
                free($7);
            } modrm immediate 
           ;

instruction_list: instruction_list instruction {
             /* if immediate_flag = true, the length of immediate has not been
                counted. */
             int i;
             for (i = 0; i< prefix_array.size(); i++)
                print_all(prefix_array[i]);
             cout << endl;
             cout << endl;
             init_instruction();
           }
	   | {
             /* put all the initial stuff here */
             init_instruction(); 
           }
           ; 

mandatory: MANDATORY_PREFIX { $$ = $1; current_length++;}
          |
          {$$ = NULL;}
          ;
escape:	  ESCAPE { $$ = $1; current_length++;}
          |{ $$ = NULL; }
          ;	
opcode:   BIT_8 {
                 $$ = $1;
                 current_length++; 
		 /* generate prefix for this opcode */
                 init_prefix();
          }
          ;
modrm:	  MODA REG R_M {
		        mod_16_bit_address(false, "xxx");
		        mod_32_bit_address(false, "xxx");
                        /* 00, 01, 10*/
                       }
	 | MOD REG R_M {
			mod_16_bit_address(true, "xxx");
                        mod_32_bit_address(true, "xxx");
                        /* 00, 01, 10, 11*/
                       }
         | MOD BIT_3 R_M {
			mod_16_bit_address(true,(string)$2);
                        mod_32_bit_address(true,(string)$2);
                        free($2);
                        /* opcode extensions */
                        }	
	 |
	 ;
immediate: IMM8 { current_length++; }
	 | IMMEDIATE { 
			immediate_flag = true;
                     }
	 | IMMEDIATE ',' IMM8 { immediate_flag = true;current_length++; }
	 | IMMEDIATE ',' IMM16 { immediate_flag = true;current_length+=2; }
	 | IMM16 ',' IMM8 { current_length+=3; }
         |
         ;
	
%%

void
init_instruction()
{
               prefix_array.clear();
               mod_16_array.clear();
               mod_32_array.clear();
               immediate_flag = false;
               gr1 = gr2 = gr3 = gr4 = false;
               current_string = "";
               current_length = 0;
               current_address_32 = INIT_ADDRESS_MODE; 
               current_oprand_32  = INIT_OPRAND_MODE;
}

string
removeBlanks(string bits)
{
  string temp = "";
  unsigned int i;
  for (i = 0; i < bits.size(); i++) 
    if (bits[i] != ' ') 
      temp += bits[i];
  return temp;
}

/* Table 2-1, Intel Manual */
/* I just list the combinations for 16-Bit addressing for Table 2-1*/
/* coding style is picked for easy reading and can be more efficient */
void
mod_16_bit_address(int mod_11, string ext)
{
	struct mod_ mod;
	/* mod = 00 */
         if (ext == "")
                  mod.pattern = ":00xxx0xx";
         else
                  mod.pattern = ":00"+ext+"0xx";
         mod.size = 1;
	 mod_16_array.push_back(mod);

         if (ext == "")
                  mod.pattern = ":00xxx10x";
         else
                  mod.pattern = ":00"+ext+"10x";
         mod.size = 1;
         mod_16_array.push_back(mod);

         if (ext == "")
                  mod.pattern = ":00xxx110";
         else
                  mod.pattern = ":00"+ext+"110";
         mod.size = 3;
         mod_16_array.push_back(mod);

         if (ext == "")
                  mod.pattern = ":00xxx111";
         else
                  mod.pattern = ":00"+ext+"111";
         mod.size = 1;
         mod_16_array.push_back(mod);

        /* mod = 01 */
	 if (ext == "")
                  mod.pattern = ":01xxxxxx";
         else
                  mod.pattern = ":01"+ext+"xxx";
         mod.size = 2;
         mod_16_array.push_back(mod);
        
        /* mod = 10 */
         if (ext == "")
                  mod.pattern = ":10xxxxxx";
         else
                  mod.pattern = ":10"+ext+"xxx";
         mod.size = 3;
         mod_16_array.push_back(mod);

        /* mod = 11 */
        if (mod_11 = true) {
             mod.size = 1;
             if (ext == "")
                  mod.pattern = ":11xxxxxx";
             else
                  mod.pattern = ":11"+ext+"xxx";
             mod_16_array.push_back(mod);
        }
}

/* Table 2-2, 2-3. */
void
mod_32_bit_address(int mod_11, string ext)
{
	struct mod_ mod;
	int i, j, p, q;

	string Mod[3] = {"00", "01", "10"};
	int size[3] = {1, 2, 5};
        string RM[3] = {"0xx", "101", "11x"};

	string SIB;
	string index[3] = {"0xx","101", "11x"};
	string base[4] = {"0xx","100", "11x", "101"}; 

        for (i = 0; i< 3; i++) { 
            for (j =0; j <3; j++) {
	 	if (ext == "")
                	  mod.pattern = "xxx"+RM[j];
         	else
               		  mod.pattern = ext+RM[j];
         	if (Mod[i] == "00" && RM[j]=="101") mod.size = size[2];
                else mod.size = size[i];
         	mod.pattern =":"+ Mod[i] + mod.pattern;
         	mod_32_array.push_back(mod);
            }

	/* SIB presented here */
            for (p=0; p<3; p++)
       	        for (q=0; q<4; q++) {
	           if (Mod[i] == "00")
                     SIB = ":xx"+ index[p]+ base[q];
                   else
                     SIB = ":xx"+ index[p]+ "xxx";
                   if (ext =="")
                        mod.pattern =":"+Mod[i] + "xxx100" + SIB;
                        else
              		mod.pattern =":"+Mod[i] + ext +"100" + SIB;
                   if (Mod[i] == "00") mod.size = 2;
		   if (Mod[i] == "00" && base[q] == "101") mod.size = 6;
                   if (Mod[i] == "01") mod.size = 3;
                   if (Mod[i] == "10") mod.size = 6;
	           mod_32_array.push_back(mod);
                  if (Mod[i] != "00") break; 
          }
            
       }
       if (mod_11 == true) {
           if (ext =="")
              mod.pattern =":11" + ext + "xxx";
           else
              mod.pattern =":11" + ext + "xxx";                                                                                               
           mod.size = 1;
           mod_32_array.push_back(mod);
       }
                                                                                                      
}

void
print_all(string prefix_string)
{
	int i, j, full_imm_size = 0;
        string temp_string;
        string w;
        string::size_type pos;
        vector<struct mod_>  *mod_array;

        /* save global variables, address mode and oprand mode must be saved */

        int saved_current_address_32 = current_address_32;
        int saved_current_oprand_32 = current_oprand_32;
        string saved_current_string = current_string;
        
        /* toggle address and oprand mode based on prefix Bytes */ 
	pos = prefix_string.find (Pre67h,0);
        if (pos != string::npos) current_address_32 = !(current_address_32);
                                                                                                      
        pos = prefix_string.find (Pre66h,0);
        if (pos != string::npos) current_oprand_32 = !(current_oprand_32);
       
        if (current_address_32 == false) mod_array = &mod_16_array;
        else mod_array = &mod_32_array;

        pos = current_string.find ("w",0);
        if (pos != string::npos) 
            for (i=0; i<2; i++){
	    	if (i == 0) w = "0";
            	if (i == 1) w = "1";
            	if (immediate_flag == false) w = "x";
            	current_string.replace(pos, 1, w);
          	if (immediate_flag && w =="0") full_imm_size = 1;
          	if (!current_oprand_32 && immediate_flag && w =="1")
                     full_imm_size = 2;
          	if (current_oprand_32 && immediate_flag && w =="1")
                     full_imm_size = 4;
        	for (j=0; j<(*mod_array).size(); j++)
                          cout
                            << current_ID <<  "\t"
                            << prefix_string + current_string
                                + (*mod_array)[j].pattern
                            << "\t"
                            << prefix_string.length()/8 + current_length
                               + (*mod_array)[j].size + full_imm_size
                            << endl;
               if ((*mod_array).size() == 0)
                       cout
                            << current_ID << "\t"
                            << prefix_string + current_string << "\t"
                            << prefix_string.length()/8 + current_length
                                + full_imm_size
                            << endl;
               if (immediate_flag == false) break; /* only do once */
        }
        /* no "w" found */
        else {
            if (!current_oprand_32 && immediate_flag) full_imm_size = 2;
            if (current_oprand_32 && immediate_flag) full_imm_size = 4;
            for (j=0; j<(*mod_array).size(); j++)
                        cout
                            << current_ID << "\t"
                            << prefix_string + current_string
                                 + (*mod_array)[j].pattern << "\t"
                            << prefix_string.length()/8 + current_length
                               + (*mod_array)[j].size + full_imm_size
                            << endl;
            if ((*mod_array).size() == 0)
                       cout
                            << current_ID << "\t"
                            << prefix_string + current_string << "\t"
                            << prefix_string.length()/8 + current_length
                               + full_imm_size
                            << endl;
      }        
      current_oprand_32 = saved_current_oprand_32;
      current_address_32 = saved_current_address_32;
      current_string = saved_current_string;
}

void
save_permutation(vector<string>& perm)
{
  string new_p;
  for(vector<string>::iterator it= perm.begin();it!=perm.end();++it) 
    new_p = new_p + *it + ":"; 
  
  prefix_array.push_back(new_p);
}

void
init_prefix()
{
	int i, j, k, l;
        int p, q, r;
	string new_prefix;
	vector<string> v_gr1, v_gr2, v_gr3, v_gr4;
	vector<string> perm_set, cnext, pnext;

	/* I will modify the instruction format to give prefix group information */
        /* in the future, so I can add a prefix filter for each instruction */
	if (gr1 == false)
	 for (p=0; p<3; p++)
	    if (Pre_gr1[p]!= "")
		v_gr1.push_back(Pre_gr1[p]);

        if (gr2 == false)
	for (p=0; p<8;p++)
            if (Pre_gr2[p]!= "")
                v_gr2.push_back(Pre_gr2[p]);

        if (gr3 == false)
	for (p=0; p<1;p++)
            if (Pre_gr3[p]!= "")
                v_gr3.push_back(Pre_gr3[p]);

        for (p=0; p<1;p++)
            if (Pre_gr4[p]!= "")
                v_gr4.push_back(Pre_gr4[p]);


	/* 0 prefix, 1 prefix,...4 prefix */

        /* 0 prefix */
        prefix_array.push_back(new_prefix);

	/* at least one prefix */
        for (i=0; i<=v_gr1.size();i++) {	
	   if (i!= v_gr1.size()) perm_set.push_back(v_gr1[i]);		
	   for 	(j=0; j<=v_gr2.size();j++) {
	      if (j!= v_gr2.size()) perm_set.push_back(v_gr2[j]);	
	      for (k=0; k<=v_gr3.size();k++) {
                  if (k!= v_gr3.size()) perm_set.push_back(v_gr3[k]);
	          for (l= 0; l<=v_gr4.size();l++){
		     if (l!= v_gr4.size())
                            perm_set.push_back(v_gr4[l]);
                     if (perm_set.size())
                            prefix_array_combination.push_back(perm_set);
                     if (l!= v_gr4.size())
                            perm_set.erase(perm_set.end()-1, perm_set.end());
                  }
                  if (k!= v_gr3.size())
                     perm_set.erase(perm_set.end()-1, perm_set.end());
             }
             if (j!= v_gr2.size())
                     perm_set.erase(perm_set.end()-1, perm_set.end());
         }
         if (i!= v_gr1.size())
                 perm_set.erase(perm_set.end()-1, perm_set.end());
	}

        for (q= prefix_array_combination.size()-1; q >=0; q--)
               vector_permutation(
                                  prefix_array_combination[q],
                                  pnext,
                                  save_permutation);
        v_gr1.clear();
        v_gr2.clear();
	v_gr3.clear();	
	v_gr4.clear();
        for (q= prefix_array_combination.size()-1; q >=0; q--)
          prefix_array_combination[q].clear();
        prefix_array_combination.clear();
}


int
main(int argc, char **argv) 
{
	FILE *in;
	if ( argc > 1 ) {
          in = fopen( argv[1], "r" );
          if (in == NULL) {
            printf("File Name Error! \n");
            return -1;
          }
        }
        else {
           printf("Format: command instructionFormatFile \n");
           return -1;
        }
        yyin = in;
	lex_init();
	yyparse();
	return 0;
}
